﻿using System.Configuration;
using System.Diagnostics.CodeAnalysis;
using UDFS.App.Forms;
using UDFS.App.Services.Connection;
using UDFS.App.Services.Files;
using UDFS.App.Services.Functions;
using UDFS.App.Services.Logging;
using UDFS.App.Stores.Connection;
using UDFS.App.Stores.Functions;
using UDFS.Shared.Services;

namespace UDFS.App
{
    /// <summary>
    /// Class for starting application.
    /// </summary>
    public class Bootstrapper
    {
        #region Constants
        // Directories
        private const string APPDATA_DIR_NAME = "UDFSApp";
        private const string BACKUPS_DIR_NAME = "BACKUP_FILES";
        private const string LOGS_DIR_NAME = "LOG_FILES";
        private const int BACKUP_LIFETIME_DAYS = 30;

        // Configuration keys
        private const string UDFS_DIR_KEY = "UDFSDir";
        private const string BACKUPS_DIR_KEY = "BackupFilesDir";
        private const string LOGS_DIR_KEY = "LogFilesDir";
        private const string BACKUP_LIFETIME_DAYS_KEY = "BackupFileLifetimeDays";
        #endregion

        #region Directories
        private DirectoryInfo _appDir;
        private DirectoryInfo _appdataDir;
        private DirectoryInfo _backupFilesDir;
        private DirectoryInfo _logFilesDir;
        private DirectoryInfo _udfsDir;
        #endregion

        #region Settings
        private int _backupLifetimeDays = 30;
        #endregion

        #region Stores
        private IConnectionStore _connStore;
        private IFunctionsStore _functionsStore;
        #endregion

        #region Services
        private ISetConnectionService _setConnectionService;
        private IRunFunctionService _runFunctionService;
        private ILoadFunctionsService _loadFunctionsService;
        private ILogService _logService;
        private IClearFilesService _clearBackupFilesService;
        #endregion

        public Bootstrapper()
        {
            InitDirectories();
            InitSettings();
            InitStores();
            InitServices();
        }

        public void Run()
        {
            _setConnectionService.SetConnection();
            _loadFunctionsService.Load();
            _clearBackupFilesService.Clear();
            Application.Run(new MainForm(_connStore, _functionsStore, _setConnectionService, _loadFunctionsService));
        }

        /// <summary>
        /// Initializes directories.
        /// </summary>
        [MemberNotNull(nameof(_appDir))]
        [MemberNotNull(nameof(_appdataDir))]
        [MemberNotNull(nameof(_backupFilesDir))]
        [MemberNotNull(nameof(_logFilesDir))]
        [MemberNotNull(nameof(_udfsDir))]
        private void InitDirectories()
        {
            _appDir = new DirectoryInfo(AppDomain.CurrentDomain.BaseDirectory);
            _appdataDir = new DirectoryInfo(Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData), APPDATA_DIR_NAME));

            string? udfsDir = ConfigurationManager.AppSettings[UDFS_DIR_KEY];
            if (string.IsNullOrWhiteSpace(udfsDir)) udfsDir = _appDir.FullName;
            _udfsDir = new DirectoryInfo(udfsDir);

            string? backupsDir = ConfigurationManager.AppSettings[BACKUPS_DIR_KEY];
            if (string.IsNullOrWhiteSpace(backupsDir)) backupsDir = Path.Combine(_appdataDir.FullName, BACKUPS_DIR_NAME);
            _backupFilesDir = new DirectoryInfo(backupsDir);

            string? logsDir = ConfigurationManager.AppSettings[LOGS_DIR_KEY];
            if (string.IsNullOrWhiteSpace(logsDir)) logsDir = Path.Combine(_appdataDir.FullName, LOGS_DIR_NAME);
            _logFilesDir = new DirectoryInfo(logsDir);
        }
        /// <summary>
        /// Initializes settings.
        /// </summary>
        [MemberNotNull(nameof(_backupLifetimeDays))]
        private void InitSettings()
        {
            string? backupLifetimeDays = ConfigurationManager.AppSettings[BACKUP_LIFETIME_DAYS_KEY];
            if (!int.TryParse(backupLifetimeDays, out _backupLifetimeDays)) _backupLifetimeDays = BACKUP_LIFETIME_DAYS;
        }
        /// <summary>
        /// Intializes stores.
        /// </summary>
        [MemberNotNull(nameof(_connStore))]
        [MemberNotNull(nameof(_functionsStore))]
        private void InitStores()
        {
            _connStore = new ConnectionStore();
            _functionsStore = new FunctionsStore();
        }
        /// <summary>
        /// Initializes services.
        /// </summary>
        [MemberNotNull(nameof(_setConnectionService))]
        [MemberNotNull(nameof(_runFunctionService))]
        [MemberNotNull(nameof(_loadFunctionsService))]
        [MemberNotNull(nameof(_logService))]
        [MemberNotNull(nameof(_clearBackupFilesService))]
        private void InitServices()
        {
            _setConnectionService = new FormSetConnectionService(_connStore);
            _runFunctionService = new RunFunctionService(_connStore, _logFilesDir, _backupFilesDir);
            _loadFunctionsService = new DirLoadFunctionService(_udfsDir, _functionsStore, _runFunctionService);
            _logService = new FileLogService(_logFilesDir);
            _clearBackupFilesService = new ClearBackupFilesService(_backupFilesDir, _backupLifetimeDays);
        }
    }
}
