﻿using System.Data;
using UDFS.Shared.Database;
using UDFS.Shared.Events;

namespace UDFS.App.Database
{
    /// <summary>
    /// Abstract base class of transactional database.
    /// </summary>
    internal abstract class TransactionalDatabase : IDatabase, ITransactionalDatabase
    {
        /// <summary>
        /// Database connection string.
        /// </summary>
        protected readonly string ConnectionString;
        /// <summary>
        /// Indicates if database transaction is active.
        /// </summary>
        protected bool IsTransactionActive = false;
        /// <summary>
        /// Connection to database.
        /// </summary>
        private IDbConnection? _connection;
        /// <summary>
        /// Database transaction.
        /// </summary>
        private IDbTransaction? _transaction;

        /// <summary>
        /// Initializes a new instance of <see cref="TransactionalDatabase"/> class.
        /// </summary>
        /// <param name="connectionString"></param>
        public TransactionalDatabase(string connectionString)
        {
            ConnectionString = connectionString;
        }

        /// <summary>
        /// Occurs when system message is reported.
        /// </summary>
        public event ReportEventHandler? SysMessageReported;
        /// <summary>
        /// Occurs when user message is reported.
        /// </summary>
        public event ReportEventHandler? UsrMessageReported;

        /// <summary>
        /// Connection to database.
        /// </summary>
        protected IDbConnection Connection
        {
            get
            {
                if (_connection == null)
                {
                    _connection = CreateConnection();
                    ReportSystem("SPOJENÍ VYTVOŘENO");
                }
                return _connection;
            }
        }
        /// <summary>
        /// Database transaction.
        /// </summary>
        protected IDbTransaction? Transaction
        {
            get
            {
                if (_transaction == null && IsTransactionActive)
                {
                    PrepareConnection();
                    _transaction = Connection.BeginTransaction();
                    ReportSystem("TRANSAKCE ZAHÁJENA");
                }
                return _transaction;
            }
        }

        #region Connection
        protected abstract IDbConnection CreateConnection();
        protected IDbCommand CreateCommand()
        {
            IDbCommand command = Connection.CreateCommand();
            command.Connection = Connection;
            command.Transaction = Transaction;
            return command;
        }
        protected void PrepareConnection()
        {
            if (Connection.State == ConnectionState.Closed)
            {
                Connection.Open();
                ReportSystem("SPOJENÍ OTEVŘENO");
            }
        }
        protected void ReleaseConnection()
        {
            if (!IsTransactionActive && _connection != null)
            {
                _connection.Close();
                ReportSystem("SPOJENÍ UZAVŘENO");
            }
        }
        #endregion

        #region Transaction
        public void BeginTransaction()
        {
            if (_transaction != null || IsTransactionActive)
                throw new InvalidOperationException("Transaction already exists.");
            IsTransactionActive = true;
            ReportSystem("TRANSAKCE AKTIVOVÁNA");
        }
        public void CommitTransaction()
        {
            if (_transaction != null)
            {
                _transaction?.Commit();
                ReportSystem("TRANSAKCE POTVRZENA");
            }
        }
        public void RollbackTransaction()
        {
            if (_transaction != null)
            {
                _transaction.Rollback();
                ReportSystem("TRANSAKCE ODVOLÁNA");
            }
        }
        public void EndTransaction()
        {
            _transaction?.Dispose();
            _transaction = null;
            IsTransactionActive = false;
            ReportSystem("TRANSAKCE DEAKTIVOVÁNA");
            ReleaseConnection();
        }
        #endregion

        #region Commands
        public virtual int Execute(string commandText, params object[] parameters)
        {
            PrepareConnection();
            using IDbCommand command = CreateCommand();
            command.CommandText = commandText;
            foreach (object parameter in parameters)
            {
                command.Parameters.Add(parameter);
            }
            ReportUser($"Spuštěn příkaz: {command.CommandText} [{string.Join(',', parameters)}]");
            int result = command.ExecuteNonQuery();
            ReleaseConnection();
            ReportUser($"Ovlivněno {result} záznamů.");
            return result;
        }
        public virtual DataTable Select(string commandText, params object[] parameters)
        {
            PrepareConnection();
            using IDbCommand command = CreateCommand();
            command.CommandText = commandText;
            foreach (object parameter in parameters)
            {
                command.Parameters.Add(parameter);
            }
            ReportUser($"Spuštěn příkaz: {command.CommandText} [{string.Join(',', parameters)}]");
            using IDataReader reader = command.ExecuteReader();
            DataTable data = new();
            data.Load(reader);
            ReleaseConnection();
            ReportUser($"Navráceno {data.Rows.Count} záznamů.");
            return data;
        }
        public virtual T SelectItem<T>(string commandText, T defaultValue, params object[] parameters)
        {
            PrepareConnection();
            using IDbCommand command = CreateCommand();
            command.CommandText = commandText;
            foreach (object parameter in parameters)
            {
                command.Parameters.Add(parameter);
            }
            ReportUser($"Spuštěn příkaz: {command.CommandText} [{string.Join(',', parameters)}]");
            object? item = command.ExecuteScalar();
            ReleaseConnection();

            if (item == null || item == DBNull.Value)
            {
                ReportUser($"Navrácena defaultní hodnota: {defaultValue}");
                return defaultValue;
            }
            try
            {
                object itemChanged = Convert.ChangeType(item, typeof(T));
                if (itemChanged is T value)
                {
                    ReportUser($"Navrácena hodnota: {value}");
                    return value;
                }
            }
            catch { }
            ReportUser($"Navrácena defaultní hodnota: {defaultValue}");
            return defaultValue;
        }
        public virtual List<T> SelectList<T>(string commandText, T defaultValue, params object[] parameters)
        {
            PrepareConnection();
            using IDbCommand command = CreateCommand();
            command.CommandText = commandText;
            foreach (object parameter in parameters)
            {
                command.Parameters.Add(parameter);
            }
            ReportUser($"Spuštěn příkaz: {command.CommandText} [{string.Join(',', parameters)}]");
            using IDataReader reader = command.ExecuteReader();
            List<T> list = new();
            while (reader.Read())
            {
                object item = reader.GetValue(0);
                if (item == null || item == DBNull.Value) list.Add(defaultValue);
                else if (item is T value) list.Add(value);
                else list.Add(defaultValue);
            }
            ReleaseConnection();
            ReportUser($"Navráceno {list.Count} záznamů.");
            return list;
        }
        #endregion

        public void Dispose()
        {
            _transaction?.Dispose();
            _transaction = null;
            ReportSystem("TRANSAKCE UVOLNĚNA");
            _connection?.Dispose();
            _connection = null;
            ReportSystem("SPOJENÍ UVOLNĚNO");
            GC.SuppressFinalize(this);
        }
        /// <summary>
        /// Reports system message.
        /// </summary>
        /// <param name="message">Message to report.</param>
        protected void ReportSystem(string message)
        {
            SysMessageReported?.Invoke(this, new ReportEventArgs(message));
        }
        /// <summary>
        /// Reports user message.
        /// </summary>
        /// <param name="message">Message to report.</param>
        protected void ReportUser(string message)
        {
            UsrMessageReported?.Invoke(this, new ReportEventArgs(message));
        }
    }
}
