﻿using System.Reflection;
using UDFS.App.Stores.Functions;
using UDFS.Shared.Controls;
using UDFS.Shared.Services;

namespace UDFS.App.Services.Functions
{
    /// <summary>
    /// Service for loading functions from directory.
    /// </summary>
    internal class DirLoadFunctionService : ILoadFunctionsService
    {
        private const string LIB_FILE_PREFIX = "UDFS.Functions.";
        private const string LIB_FILE_EXTENSION = ".dll";

        private readonly DirectoryInfo _udfsDir;
        private readonly IFunctionsStore _udfsControlsStore;
        private readonly IRunFunctionService _runFunctionService;

        /// <summary>
        /// Intializes a new instance of <see cref="DirLoadFunctionService"/> class.
        /// </summary>
        /// <param name="udfsDir">Functions directory.</param>
        /// <param name="udfsControlsStore">Function controls store.</param>
        /// <param name="runFunctionService">Service for running functions.</param>
        public DirLoadFunctionService(DirectoryInfo udfsDir, IFunctionsStore udfsControlsStore, IRunFunctionService runFunctionService)
        {
            _udfsDir = udfsDir;
            _udfsControlsStore = udfsControlsStore;
            _runFunctionService = runFunctionService;
        }

        public void Load()
        {
            try
            {
                _udfsControlsStore.Controls.Clear();
                _udfsDir.Create();

                foreach (var lib in _udfsDir.GetFiles($"{LIB_FILE_PREFIX}*{LIB_FILE_EXTENSION}", SearchOption.AllDirectories))
                {
                    Assembly assembly = Assembly.LoadFrom(lib.FullName);
                    if (assembly == null) continue;
                    foreach (var udfControl in assembly.GetExportedTypes().Where(x => typeof(IFunctionControl).IsAssignableFrom(x) && x.IsClass && !x.IsAbstract))
                    {
                        try
                        {
                            if (Activator.CreateInstance(udfControl) is not IFunctionControl userFunction) continue;
                            userFunction.Initialize(_runFunctionService);
                            _udfsControlsStore.Controls.Add(userFunction);
                        }
                        catch (Exception ex)
                        {
                            MessageBox.Show($"Nepodařilo se načíst uživatelskou funkci {udfControl}:\n{ex.Message}");
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show($"Načítání funkcí selhalo:\n{ex.Message}");
            }
        }
    }
}
