﻿using UDFS.App.Database;
using UDFS.App.Forms;
using UDFS.App.Services.Logging;
using UDFS.App.Stores.Connection;
using UDFS.Shared.Database;
using UDFS.Shared.Functions;
using UDFS.Shared.Services;

namespace UDFS.App.Services.Functions
{
    /// <summary>
    /// Service for running functions.
    /// </summary>
    internal class RunFunctionService : IRunFunctionService
    {
        private readonly IConnectionStore _connStore;
        private readonly DirectoryInfo _logsDir;
        private readonly DirectoryInfo _backupsDir;

        public RunFunctionService(IConnectionStore connStore, DirectoryInfo logsDir, DirectoryInfo backupsDir)
        {
            _connStore = connStore;
            _logsDir = logsDir;
            _backupsDir = backupsDir;
        }

        public async void Run(IUDFunction function)
        {
            SetFormsEnabled(false);

            ProgressForm progressForm = new(function.FunctionName);
            ILogService logService = new FileLogService(_logsDir);

            logService.LogChanged += () => progressForm.RefreshLog(logService.Log);
            progressForm.CancelRequested += () =>
            {
                function.IsCancelPending = true;
                logService.LogSystem("PŘIJATA ŽÁDOST O UKONČENÍ FUNKCE");
            };
            progressForm.Show();
            await Task.Run(() => RunInternal(function, logService));

            logService.Save();
            progressForm.SetStateDone();
            if (progressForm.IsCancelled) progressForm.SetStateCancelled();
            else progressForm.SetStateDone();
            SetFormsEnabled(true);
        }
        /// <summary>
        /// Runs specified function with specified log service.
        /// </summary>
        /// <param name="function"></param>
        /// <param name="logService"></param>
        private void RunInternal(IUDFunction function, ILogService logService)
        {
            try
            {
                using IDatabase database = _connStore.CreateDatabase?.Invoke() ?? throw new InvalidOperationException("Není připojena žádná databáze.");
                database.SysMessageReported += (s, e) => logService.LogSystem(e.Message);

                function.Database = database;
                function.SysMessageReported += (s, e) => logService.LogSystem(e.Message);
                function.UsrMessageReported += (s, e) => logService.LogUser(e.Message);

                IBackupableDatabase? backupableDatabase = database as IBackupableDatabase;
                ITransactionalDatabase? transactionalDatabase = database as ITransactionalDatabase;

                // Záloha
                if (function.CreateBackup && backupableDatabase == null)
                {
                    string message = "Připojená databáze neumožňuje zálohování. Přesto pokračovat?";
                    logService.LogSystem(message);
                    if (MessageBox.Show(message, "Upozornění", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.No)
                    {
                        logService.LogSystem("Ne. Akce ukončena uživatelem.");
                        throw new FunctionCancelledException();
                    }
                    logService.LogSystem("Ano. Akce potvrzena uživatelem.");
                }

                // Transakce
                if (function.UseTransaction && transactionalDatabase == null)
                {
                    string message = "Připojená databáze neumožňuje použití transakčního zpracování. Přesto pokračovat?";
                    logService.LogSystem(message);
                    if (MessageBox.Show(message, "Upozornění", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.No)
                    {
                        logService.LogSystem("Ne. Akce ukončena uživatelem.");
                        throw new FunctionCancelledException();
                    }
                    logService.LogSystem("Ano. Akce potvrzena uživatelem.");
                }

                try
                {
                    backupableDatabase?.CreateBackup(_backupsDir);
                    transactionalDatabase?.BeginTransaction();
                    function.Run();
                    transactionalDatabase?.CommitTransaction();
                }
                catch (FunctionCancelledException)
                {
                    transactionalDatabase?.RollbackTransaction();
                    logService.LogSystem("FUNKCE UKONČENA NA ŽÁDOST");
                }
                catch (Exception)
                {
                    transactionalDatabase?.RollbackTransaction();
                    throw;
                }
                finally
                {
                    transactionalDatabase?.EndTransaction();
                }
            }
            catch (Exception ex)
            {
                logService.LogSystem("V PRŮBĚHU FUNKCE VYVSTALA CHYBA");
                logService.LogSystem(ex.Message);
                logService.LogSystem(ex.StackTrace ?? "");
            }
        }
        /// <summary>
        /// Sets enabled value to all opened forms of application.
        /// </summary>
        /// <param name="enabled"></param>
        private static void SetFormsEnabled(bool enabled)
        {
            foreach (Form form in Application.OpenForms) form.Enabled = enabled;
        }
    }
}
